module Kristall_Modul

! Allgemeine Parameter
integer,parameter::DIM=2                 ! Raumdimension (2D oder 3D)
integer,parameter::AUSGEBEN_ALLE=500
real,parameter::   DELTA_T=0.0002           ! Zeitschrittweite 
real,parameter::   T_MAX=50.0
integer,parameter::NX=40             
integer,parameter::NY=20
integer,parameter::NXX=10             
integer,parameter::NYY=10
integer,parameter::STARTX=0             
integer,parameter::STARTY=NY+NYY/2+10
real,parameter::VX=0
real,parameter::VY=0
! Physikalische Konstanten
real,parameter::SIGMA=1.0           
real,parameter::EPS=0.4           
real,parameter::MASSE1=1.0                   
real,parameter::MASSE2=2.0
real,parameter::TEMP=0.1           
real,parameter::RGG=2**(1.0/6.0)*SIGMA
real,parameter::ERDBESCHLEUNIGUNG=5
real,parameter::WUNSCHTEMPERATUR=TEMP
real,parameter::GEDAEMPFTE_ZEIT=10
real,parameter::FALL_START=10
  
integer,parameter::SPIEGELLINKS=-NX*RGG/2 
integer,parameter::SPIEGELRECHTS=NX*RGG/2
integer,parameter::SPIEGELUNTEN=-NY*RGG/2
! Eigene Datentypen
type Atom                                ! Himmelskoerper:
  real::m                                !   Masse
  real,dimension(DIM)::x,v,F,F_alt             !   Ort,Geschwindigkeit,Kraft
  integer::typ                           !   Typ (fuer Darstellung)
  integer::zelle
end type

type Atome                               ! Sonnensystem:
  integer::N                             !   Anzahl der Himmelskoerper
  type(Atom),dimension(:),pointer::a     !   Feld der Himmelskoerper
end type

integer,parameter::NACHBARN=3**DIM
logical,parameter::ABSCHNEIDEN=.true.
real,parameter::R_AB=2.5*SIGMA

integer,parameter::NGX=NX+5
integer,parameter::NGY=2*(1.2*STARTY) 

type Gebiet
  integer::N,Nx,Ny
  real::x0,y0
  real::d
  integer,dimension(:,:),pointer::Nachbar
  integer,dimension(:),pointer::Anzahl,Start
end type

contains

! ***  Gebietsaufteilung  *****************************************************

subroutine Aktualisiere_AtomZelle(G,A)
      type(Atome)::A
      type(Gebiet)::G
  integer::i
  do i=1,A%N
    A%a(i)%zelle=Zellnummer(G,A,i)
  end do
end subroutine

function Zellnummer(G,A,i)
      integer::Zellnummer; type(Gebiet)::G; type(Atome)::A; integer::i
  integer::zx,zy
  zx=int((A%a(i)%x(1)-G%x0)/G%d)
  zy=int((A%a(i)%x(2)-G%y0)/G%d)
  if ( (0<=zx) .and. (zx<G%Nx) .and. (0<=zy) .and. (zy<G%Ny) ) then
    Zellnummer=1+zx+zy*G%Nx
  else
    Zellnummer=0
  end if
end function

function Zellen_benachbart(G,i,j)
      logical::Zellen_benachbart; type(Gebiet)::G; integer::i,j
  integer::ix,iy,jx,jy
  ix=mod((i-1),G%Nx); iy=int((i-1)/G%Nx)
  jx=mod((j-1),G%Nx); jy=int((j-1)/G%Nx)
  if ( (i/=0) .and. (j/=0) .and. (abs(ix-jx)<=1) .and. (abs(iy-jy)<=1) ) then
    Zellen_benachbart=.true.
  else
    Zellen_benachbart=.false.
  end if
end function

subroutine Atome_sortieren(G,A)
      type(Gebiet)::G; type(Atome)::A
  integer::i,k,l,z,zn
  type(Atom)::aa 
  ! Anzahl und Startposition fuer jedes Zelle bestimmen
  G%Anzahl(:)=0
  do k=1,A%N
    i=A%a(k)%zelle 
    G%Anzahl(i)=G%Anzahl(i)+1
  end do
  G%Start(0)=1
  do i=1,G%N
    G%Start(i)=G%Start(i-1)+G%Anzahl(i-1)
  end do
  ! Atome einsortieren
  G%Anzahl(:)=0    ! Zahl der richtig einsortierten Atome
  z=0              ! Aktuelle Zelle
  do while (z<G%N)      ! "<", da letzte Zelle automatisch richtig
    ! Solange noch Teilchen in Zelle z fehlen, durch Tauschen welche suchen
    do while ( G%Start(z)+G%Anzahl(z)<G%Start(z+1) )
      k=G%Start(z)+G%Anzahl(z)   ! erstes potentiell falsche Atom in Zelle z
      zn=A%a(k)%zelle            ! Zelle, wo das aktuelle Atom hin muss
      do while (zn/=z)
        ! aktuelles Atom richtig einsortieren und dortiges Atom herholen
        l=G%Start(zn)+G%Anzahl(zn)   ! hier muss das aktuelle Atom hin
        G%Anzahl(zn)=G%Anzahl(zn)+1  ! und hier ist es dann auch richtig
        aa=A%a(l); A%a(l)=A%a(k); A%a(k)=aa  ! Atome k und l tauschen
        zn=A%a(k)%zelle              ! Gebiet des geholten Atoms
      end do
      G%Anzahl(z)=G%Anzahl(z)+1  ! aktuelles Teilchen ist richtig einsortiert
    end do
    z=z+1               ! aktuelle Zelle fertig, also zur naechsten
  enddo
  ! letze Zelle wurde nicht ganz durchlaufen, deshalb noch Anzahl berechnen
  G%Anzahl(G%N)=A%N-G%Start(G%N)+1
end subroutine

subroutine Zellen_initialisieren(G)
      type(Gebiet)::G
  integer::i,j,k
  ! Zelleninitialisieren, Zelle 0 fr verlorene Teilchen
  G%x0=-0.5*NGX
  G%y0=-0.5*NGY
  G%d =R_AB
  G%Nx=int(NGX*RGG/G%d)+1
  G%Ny=int(NGY*RGG/G%d)+1
  G%N=G%Nx*G%Ny           ! Zahl der echten Zellen
  allocate (G%Anzahl(0:G%N))
  allocate (G%Start(0:G%N))
  allocate (G%Nachbar(0:G%N,NACHBARN))
  G%Nachbar(:,:)=-1       ! evtl. nicht vorhandene Nachbarn durch -1 markieren
  do i=0,G%N
    k=0                   ! Zahl der Nachbarzellen von Zelle i
    do j=0,G%N
      if(Zellen_benachbart(G,i,j)) then
        k=k+1
        G%Nachbar(i,k)=j  ! k-ter Nachbar von Zelle i ist j
      end if
    end do
  end do
end subroutine

! ***  Zeitintegration und Ausgabe  *******************************************

subroutine Ausgabe(A,t)
      type(Atome)::A
      real::t
  integer::i
  write(unit=1,fmt=*) A%N
  write(unit=1,fmt=*) t," fs"
  do i=1,A%N
    write(unit=1,fmt=*) A%a(i)%typ, A%a(i)%x, 0, 0
  end do
end subroutine 

subroutine Kraft(A,i,j)
      type(Atome)::A; integer::i,j
  real,dimension(DIM)::xij,Fij
  real::rr
  integer::d
  real::sigma_r_6
  xij=A%a(j)%x-A%a(i)%x                  ! Abstandsvektor von i nach j
  rr=sum(xij**2)
  sigma_r_6=SIGMA**6/(rr**3)
  
  Fij=24*EPS/rr* sigma_r_6  *(1-2*(sigma_r_6)) *xij     ! Kraft von j auf i
  A%a(i)%F=A%a(i)%F+Fij
  A%a(j)%F=A%a(j)%F-Fij
end subroutine

subroutine Berechne_F(G,A)
      type(Gebiet)::G
      type(Atome)::A
  integer::i
  integer::j
  integer::k
  integer::l
  integer::GebietDesAtoms
  do i=1,A%N
    A%a(i)%F_alt = A%a(i)%F
    A%a(i)%F = 0
  end do
  
  do k=1,A%N
    GebietDesAtoms=A%a(k)%zelle
    do j=1,NACHBARN
      if (G%Nachbar(GebietDesAtoms,j) /= -1) then
        do l=G%Start(G%Nachbar(GebietDesAtoms,j)),G%Start(G%Nachbar(GebietDesAtoms,j))+G%Anzahl(G%Nachbar(GebietDesAtoms,j))-1
          if (k < l) call Kraft(A,k,l)
        end do
      end if
    end do
  end do
end subroutine

subroutine Berechne_G(A,t)
      real::t
      type(Atome)::A
  integer::k
  do k=1,A%N
    if (t > FALL_START .or. A%a(k)%typ == 1) then
      A%a(k)%F(2)=A%a(k)%F(2)-A%a(k)%m * ERDBESCHLEUNIGUNG
    end if
  end do
  !do i=1,G%N
  !  do j=1,NACHBARN
  !    !if (G%Nachbar(i,j) /= -1) then
  !      do k=G%Start(i),G%Anzahl(i)+G%Start(i)-1
  !        do l=G%Start(G%Nachbar(i,j)),G%Anzahl(G%Nachbar(i,j))+G%Start(G%Nachbar(i,j))-1
  !          call Kraft(A,k,l)
  !        end do
  !      end do
  !    !end if
  !  end do
  !end do
  

  !do i=1,A%N
  !  do j=i+1,A%N
  !    call Kraft(A,i,j)
  !  end do
  !end do
end subroutine

subroutine Berechne_x(A,dt)
      type(Atome)::A; real::dt
  integer::i
  do i=1,A%N
    A%a(i)%x=A%a(i)%x+dt*A%a(i)%v + dt**2*A%a(i)%F/(2*A%a(i)%m)
  end do

  ! Spiegelung
  do i=1,A%N
    if (A%a(i)%x(2) < SPIEGELUNTEN) then
      A%a(i)%x(2) = 2 * SPIEGELUNTEN - A%a(i)%x(2)
      A%a(i)%v(2) = -A%a(i)%v(2)
    end if

    if (A%a(i)%x(1) > SPIEGELRECHTS) then
      A%a(i)%x(1) = 2 * SPIEGELRECHTS - A%a(i)%x(1)
      A%a(i)%v(1) = - A%a(i)%v(1)
    end if

    if (A%a(i)%x(1) < SPIEGELLINKS) then
      A%a(i)%x(1) = 2 * SPIEGELLINKS - A%a(i)%x(1)
      A%a(i)%v(1) = -A%a(i)%v(1)
    end if
  end do
  
end subroutine

subroutine Berechne_v(A,dt)
      type(Atome)::A; real::dt
  integer::i  
  do i=1,A%N
    A%a(i)%v=A%a(i)%v+dt*(A%a(i)%F + A%a(i)%F_alt)/(2*A%a(i)%m) 
  end do
end subroutine

function BerechneTemp(A)
      real::BerechneTemp
      type(Atome)::A
  real::Summe
  integer::i
  Summe=0
  do i=1,A%N
    Summe=Summe+A%a(i)%m*sum(A%a(i)%v**2)/2
  end do
  BerechneTemp=Summe/(A%N)
end function

subroutine PasseTempAn(A,WunschTemp)
      type(Atome)::A
      real::WunschTemp
  real::Temp
  real::f
  integer::i
  Temp=BerechneTemp(A)
  f=sqrt(WunschTemp/Temp)
  do i=1,A%N
    A%a(i)%v=A%a(i)%v*f 
  end do 
end subroutine

subroutine Zeitintegration(G,A,dt,t_end)
      type(Gebiet)::G
      type(Atome)::A; real::dt,t_end
  integer::k
  real::t
  t=0
  k=0                                    ! Zaehler fuer Ausgabe
  call Aktualisiere_AtomZelle(G,A) 
  call Atome_sortieren(G,A)
  call Berechne_F(G,A)
  do while(t<t_end)
    t=t+dt
    call Berechne_x(A,dt)
    call Aktualisiere_AtomZelle(G,A)
    call Atome_sortieren(G,A)
    call Berechne_F(G,A)
    call Berechne_G(A,t)
    call Berechne_v(A,dt) 
    if (t < GEDAEMPFTE_ZEIT) call PasseTempAn(A,WUNSCHTEMPERATUR)
    k=k+1
    if(mod(k,AUSGEBEN_ALLE)==0) call Ausgabe(A,t)
  end do
end subroutine

! ***  Startkonfiguration  ****************************************************

function NV_Zufallszahl()
      real::NV_Zufallszahl
  real::a1,a2,r
  r=0
  do while ((r==0).or.(1<=r))
    call random_number(a1); a1=2*a1-1
    call random_number(a2); a2=2*a2-1
    r=a1**2+a2**2
  end do
  NV_Zufallszahl=a1*sqrt(-2*log(r)/r)
end function

subroutine Temperatur(A, T)
      type(Atome)::A
      real::T
  integer::i,d
  real::s,z
  do i=1,A%N
    s=sqrt(T/A%a(i)%m)
    do d=1,DIM
      z=NV_Zufallszahl()*s
      A%a(i)%v(d)=A%a(i)%v(d)+z
    end do
  end do
end subroutine

subroutine Startkonfiguration(A)
      type(Atome)::A
  integer::i,j
  real::xShift1
  real::yShift1
  real::xShift2
  real::yShift2
  integer::offset
  ! Partikel initialisieren
  xShift1=NX*RGG/2
  yShift1=NY*RGG/2
  A%N=NX*NY + NXX*NYY
  allocate(A%a(A%N))
  offset=1
  do i=1,NX
    do j=1,NY
      !A%a(offset)%typ=1; A%a(offset)%m=MASSE1; A%a(offset)%x=(/(i-1+mod(j,2)/2.0)*RGG-xShift1,0.866*j*RGG-yShift1/); 
      A%a(offset)%typ=1; A%a(offset)%m=MASSE1; A%a(offset)%x=(/(i-1)*RGG-xShift1,j*RGG-yShift1/); 
      A%a(offset)%v=(/0.0,0.0/);
      offset=offset+1
    end do 
  end do 

  ! Kleinerer Koerper
  xShift2=NXX*RGG/2
  yShift2=NYY*RGG/2
  do i=1,NXX
    do j=1,NYY
      !A%a(offset)%typ=2; A%a(offset)%m=MASSE2; A%a(offset)%x=(/(i-1+mod(j,2)/2.0)*RGG+STARTX-xShift2,0.866*j*RGG+STARTY-yShift2/); 
      A%a(offset)%typ=2; A%a(offset)%m=MASSE2; A%a(offset)%x=(/(i-1)*RGG+STARTX-xShift2,j*RGG+STARTY-yShift2/); 
      A%a(offset)%v=(/VX,VY/);
      offset=offset+1
    end do
  end do
  call Temperatur(A,TEMP)
end subroutine

end module

! ***  Hauptprogramm  *********************************************************

program Kristall

use Kristall_Modul

type(Atome)::A
type(Gebiet)::G

open(unit=1,file="g_tropfen.xyz",action="write",status="replace")
call Zellen_initialisieren(G)
call Startkonfiguration(A)
print *, BerechneTemp(A)
call Ausgabe(A,0.0)

call Zeitintegration(G,A,DELTA_T,T_MAX)

close(unit=1)
deallocate(A%a)
deallocate(G%Anzahl)
deallocate(G%Start)
deallocate(G%Nachbar)

end program 
